## Random Sampling Function Script File ##

# library(tidyverse)   # functions to work with tibbles
# library(lubridate)   # makes working with dates easier


source(paste0(rstudioapi::getActiveProject(),"/R Functions/function_predictionModels.R"))

geomean = function(x, na.rm=TRUE){
  exp(sum(log(x[x > 0]), na.rm=na.rm) / length(x))
}

AqFishSamplePairing <- function(aqData, fishData, bylocationColumn=NULL, aqOverlapYrs=5, stdzFishLength=350){

  # FUNCTION VARIABLES
  # aqData           - dataframe of Aq   data containing Result, SampleDate, Unit
  # fishData         - dataframe of Fish data containing Result, TLAvgLength, SampleDate, Unit
  # bylocationColumn - column of location names. LOcation must be shared between Fish & Aq to be paired
  # aqStat           - Name of statistic to summarize aq data; 'mean', 'geomean', or 'median'
  # aqOverlapYrs     - number years prior Fish sample available to draw aq random sample
  # stdzFishLength  - length to determine Standardized Fish [MeHg]
  
  # bylocationColumn <- quo(Subarea) - to test function line-wise
  locationColumn <- enquo(bylocationColumn)

  # FISH DATA PREP
  fishPool <- fishData %>% 
    mutate(Year = year(SampleDate)) 
    
  # AQ DATA PREP
  aqPool <- aqData %>% 
    mutate(Year = year(SampleDate))
   
           
  # START PAIRING
  Locations <- fishPool %>% distinct(!!locationColumn) %>% pull
 
  FishAqResults <- NULL
  for(location in Locations){
    # location <- Locations[1]   # - to test function line-wise
    
    fishPool_loc <- fishPool %>% 
      filter(!!locationColumn == location)
    
    Years <- fishPool_loc %>% distinct(Year) %>% pull
    
    tempFishAqResults <- NULL
    for(yr in Years){
      # yr <- Years[1]   # - to test function line-wise
      
      ## Gather Fish Samples ##
      fishPool_locYr <- fishPool_loc %>% 
        filter(Year == yr)
      #data check#
      if(nrow(fishPool_locYr) == 1){next}
      
      # Calc Standardized Fish [MeHg]
      #if prediction requires extrapolation, make log model first (so selected with distinct(). Under ideal/optimized eating conditions a linear model would be appropriate but assume less than ideal environmental consumption conditions such that MeHg accumulation rate decreases as fish gets bigger/older.
      tempFishConcs  <- predictionModels(fishPool_locYr, .knownYcol=Result, .knownXcol=TLAvgLength, knownX=stdzFishLength, preferExtrapModel='log') %>% 
        rename(!!paste0("fish_Stdz",stdzFishLength) := predictedY,
               fish_n = n) %>% 
        filter(!grepl('gam', ModelName, ignore.case=T)) %>%   # remove 'gam' because regression model can increase then decrease as length goes up, which is unlikely
        filter( if(all(Extrapolated)) ModelName == 'log' else ModelName != 'doesnt matter') %>%  #if fish_Stdz requires extrapolation, assume log model. Under ideal/optimized eating conditions a linear model would be appropriate but assume less than ideal environmental consumption conditions such that MeHg accumulation rate decreases as fish gets bigger/older.
        filter(RMSE == min(RMSE)) %>%
        ungroup
    
      tempFishResult <- cbind(sampleLocation=location, fish_year=yr, tempFishConcs, fish_Unit=unique(fishData$Unit)) # Adds fish sample year & unit columns to StandFishConcs 
      
      
      ## Gather Aq Samples ##
      aqYrSamp <- aqPool %>% 
        filter(!!locationColumn == location,
               waterYear <= yr & waterYear > yr-aqOverlapYrs) # Includes year both fish & Aq are sampled plus aqOverlapYrs-1 previous years of Aq data
      #data check#
      # if(is.null(aqYrSamp) | nrow(aqYrSamp) < 1){next}
      
      # Calc Avg Aq Concs
      aqYrs <- paste(aqYrSamp %>% distinct(Year) %>% pull %>% sort, collapse=", ") # Collapse Aq Smple Years for storage in dataframe
      
      tempAqSamp <- tibble(aq_waterYears = aqYrs,
                           aq_n = nrow(aqYrSamp),
                           aq_Mean_pool   = mean(aqYrSamp$Result),
                           aq_Mean_wateryear = aqYrSamp %>% group_by(waterYear) %>% summarise(Avg= mean(Result)) %>% summarise(Mean=mean(Avg), .groups = 'drop') %>% pull(Mean),
                           aq_Wt.Mean     = aqYrSamp %>% group_by(waterYear) %>% summarise(Avg= mean(Result), n=n()) %>% summarise(WtMean=weighted.mean(Avg, n), .groups = 'drop') %>% pull(WtMean),
                           aq_Geomean_pool = geomean(aqYrSamp$Result),
                           aq_Geomean_wateryear = aqYrSamp %>% group_by(waterYear) %>% summarise(GeoAvg= geomean(Result)) %>% summarise(GeoMean=geomean(GeoAvg), .groups = 'drop') %>% pull(GeoMean),
                           aq_Median_pool   = median(aqYrSamp$Result),
                           aq_Median_wateryear = aqYrSamp %>% group_by(waterYear) %>% summarise(Medi= median(Result)) %>% summarise(Median=median(Medi), .groups = 'drop') %>% pull(Median),
                           aq_Unit = unique(aqData$Unit))
      
      # Store Results
      tempFishAqResults <- rbind(tempFishAqResults, cbind(tempFishResult, tempAqSamp))
     
    }
    FishAqResults <- rbind(FishAqResults, tempFishAqResults)
  }
  return(FishAqResults) 
}

