library(mgcv)       #gam()

predictionModels <- function(dataframe, .knownYcol=NULL, .knownXcol=NULL, knownX=NULL, preferExtrapModel=NULL){
  # dataframe - a AqFishSamplePairing() object
  
  if(is.null(knownX)) stop("Please include a knownX value.")
  
  # .knownYcol <- quo(Result) - to test function line-wise
  KnownYcol <- enquo(.knownYcol)
  
  # .knownXcol <- quo(TLAvgLength) - to test function line-wise
  KnownXcol <- enquo(.knownXcol)
  
  # Unique number of samples
  n <- nrow(dataframe %>% distinct(!!KnownXcol))
  
  
  # Standard Error of Regression (SER) used to determine best model
  # SER is std.dev (or measure of dispersion) around regression line
  # SER is std. deviation of predictions with dof correction, which helps select a simple model vs a more complicated model -> dof = unique(n) minus # of coefficients estimated
  # SER is also called Residual Standard Error & can be called using summary() with summary(Model)$sigma and dof using summary(model)$df[2]
  # RMSE is std. deviation of predictions without dof correction  
  # Info about how SER was calculated
  # https://www.youtube.com/watch?v=SM96OglU7Gg
  
  # GAM: Generalized additive models with integrated smoothness estimation
  iterations <- 1:(min(4, n))  # limit smoothness coefficient to k=4 or no higher than n, the number of unique samples
  # cap the smoothness of gam model at 4 or Unique number of samples, whichever is smaller
  df_gam <- NULL
  for(i in iterations){
    capture.output(    # use capture.output to stop printing cat warning
      i_gam <- try(
        dataframe %>%
          nest(data=colnames(dataframe)) %>%  # data=colnames(dataframe) needed so nest doesn't through error if dataframe is not grouped
          mutate(knownXcol = as_label(KnownXcol),
                 knownYcol = as_label(KnownYcol),
                 knownX    = knownX,
                 ModelName = paste0("gam_k=", i)) %>%
          mutate(Model = purrr::map(.x=data, .f = ~eval(parse(text=paste0("gam(", as_label(KnownYcol), " ~ s(", as_label(KnownXcol), ", k=i), data=.)"))) )) %>%
          mutate(predictedY = purrr::map_dbl(.x=Model, .f = ~predict(.x, tibble(!!KnownXcol := knownX)))) %>%
          mutate(RMSE  = purrr::map_dbl(.x=Model, .f = ~sqrt(mean((predict(.x) - dataframe %>% pull(!!KnownYcol))^2)) )) %>% 
          mutate(SER   = purrr::map_dbl(.x=Model, .f = ~sqrt( sum((predict(.x) - dataframe %>% pull(!!KnownYcol))^2) / (n - length(.x$coefficients)) ) )) 
      )
    )
    if(inherits(i_gam, "try-error")) next

    df_gam <- rbind(df_gam, i_gam)
  }
  
  
  # NLS: Nonlinear (weighted) least-squares estimates of a nonlinear model
  capture.output(    # use capture.output to stop printing cat warning
    df_nls <- try(
      dataframe %>%  # https://stackoverflow.com/a/18305916/7903538
        nest(data=colnames(dataframe)) %>%  # data=colnames(dataframe) needed so nest doesn't through error if dataframe is not grouped
        mutate(knownXcol = as_label(KnownXcol),
               knownYcol = as_label(KnownYcol),
               knownX    = knownX,
               ModelName = "nls") %>%
        mutate(Model = purrr::map(.x=data, .f = ~eval(parse(text=paste0("nls(", as_label(KnownYcol), " ~ ", as_label(KnownXcol),"^b, start=c(b=1), algorithm='plinear', data=.)"))) )) %>%
        mutate(predictedY = purrr::map_dbl(.x=Model, .f = ~predict(.x, tibble(!!KnownXcol := knownX)))) %>% 
        mutate(RMSE  = purrr::map_dbl(.x=Model, .f = ~sqrt(mean(summary(.x)$residuals^2)) )) %>% 
        mutate(SER   = purrr::map_dbl(.x=Model, .f = ~sqrt( sum(summary(.x)$residuals^2) / (n - 2) ) )) # nls() doesn't return $coefficients but it is 2 (m & b); y = m*x^b
    )
  )

  
  # LM: Fit a standard linear model y = ax + b
  df_lm <- dataframe %>%
    nest(data=colnames(dataframe)) %>%  # data=colnames(dataframe) needed so nest doesn't through error if dataframe is not grouped
    mutate(knownXcol = as_label(KnownXcol),
           knownYcol = as_label(KnownYcol),
           knownX    = knownX,
           ModelName = "lm") %>%
    mutate(Model = purrr::map(.x=data, .f = ~eval(parse(text=paste0("lm(", as_label(KnownYcol), " ~", as_label(KnownXcol),", data=.)"))) )) %>%
    mutate(predictedY = purrr::map_dbl(.x=Model, .f = ~predict(.x, tibble(!!KnownXcol := knownX)))) %>%
    mutate(RMSE  = purrr::map_dbl(.x=Model, .f = ~sqrt(mean(summary(.x)$residuals^2)) )) %>% 
    mutate(SER   = purrr::map_dbl(.x=Model, .f = ~sqrt( sum(summary(.x)$residuals^2) / (n - length(.x$coefficients)) ) ))
  
  
  # LM: Fit a Power function y = a * x^b as done by excel
  # reference: https://stackoverflow.com/a/18308359
  df_pwr <- dataframe %>%   # https://stackoverflow.com/questions/18305852/power-regression-in-r-similar-to-excel
    nest(data=colnames(dataframe)) %>%  # data=colnames(dataframe) needed so nest doesn't through error if dataframe is not grouped
    mutate(knownXcol = as_label(KnownXcol),
           knownYcol = as_label(KnownYcol),
           knownX    = knownX,
           ModelName = "pwr") %>%
    mutate(Model = purrr::map(.x=data, .f = ~eval(parse(text=paste0("lm(log(", as_label(KnownYcol), ") ~ log(", as_label(KnownXcol),"), data=.)"))) )) %>% 
    mutate(predictedY = purrr::map_dbl(.x=Model, .f = ~exp(predict(.x, tibble(!!KnownXcol := knownX)))) ) %>% 
    mutate(RMSE  = purrr::map_dbl(.x=Model, .f = ~sqrt(mean((exp(predict(.x)) - dataframe %>% pull(!!KnownYcol))^2)) )) %>% 
    mutate(SER   = purrr::map_dbl(.x=Model, .f = ~sqrt( sum((exp(predict(.x)) - dataframe %>% pull(!!KnownYcol))^2) / (n - length(.x$coefficients)) ) ))
  
  
  # LM: Fit a Logarithmic function y = a + b*ln(X) as done by excel
  # reference: https://www.statology.org/logarithmic-regression-in-r/
  df_log <- dataframe %>%
    nest(data=colnames(dataframe)) %>%  # data=colnames(dataframe) needed so nest doesn't through error if dataframe is not grouped
    mutate(knownXcol = as_label(KnownXcol),
           knownYcol = as_label(KnownYcol),
           knownX    = knownX,
           ModelName = "log") %>%
    mutate(Model = purrr::map(.x=data, .f = ~eval(parse(text=paste0("lm(", as_label(KnownYcol), " ~ log(", as_label(KnownXcol),"), data=.)"))) )) %>% 
    mutate(predictedY = purrr::map_dbl(.x=Model, .f = ~predict(.x, tibble(!!KnownXcol := knownX)))) %>% 
    mutate(RMSE  = purrr::map_dbl(.x=Model, .f = ~sqrt(mean(summary(.x)$residuals^2)) )) %>% 
    mutate(SER   = purrr::map_dbl(.x=Model, .f = ~sqrt( sum(summary(.x)$residuals^2) / (n - length(.x$coefficients)) ) ))
  
  
  # LM: Fit a Exponential function y = a * b^(X) as done by excel
  # reference: https://www.statology.org/exponential-regression-in-r/
  df_exp <- dataframe %>%
    nest(data=colnames(dataframe)) %>%  # data=colnames(dataframe) needed so nest doesn't through error if dataframe is not grouped
    mutate(knownXcol = as_label(KnownXcol),
           knownYcol = as_label(KnownYcol),
           knownX    = knownX,
           ModelName = "exp") %>%
    mutate(Model = purrr::map(.x=data, .f = ~eval(parse(text=paste0("lm(log(", as_label(KnownYcol), ") ~", as_label(KnownXcol),", data=.)"))) )) %>% 
    mutate(predictedY = purrr::map_dbl(.x=Model, .f = ~exp(predict(.x, tibble(!!KnownXcol := knownX)))) ) %>% 
    mutate(RMSE  = purrr::map_dbl(.x=Model, .f = ~sqrt(mean((exp(predict(.x)) - dataframe %>% pull(!!KnownYcol))^2)) )) %>% 
    mutate(SER   = purrr::map_dbl(.x=Model, .f = ~sqrt( sum((exp(predict(.x)) - dataframe %>% pull(!!KnownYcol))^2) / (n - length(.x$coefficients)) ) ))
  
  
  dataXcol <- dataframe %>% select(!!KnownXcol)
  
  
  if(is.null(df_gam) & inherits(df_nls, "try-error")) {
    predictionModels <- rbind(                df_lm, df_pwr, df_log, df_exp)  # exclude df_gam & df_nls
  } 
  else if(is.null(df_gam)){
    predictionModels <- rbind(        df_nls, df_lm, df_pwr, df_log, df_exp)  # exclude df_gam
  }
  else if(inherits(df_nls, "try-error")){
    predictionModels <- rbind(df_gam,         df_lm, df_pwr, df_log, df_exp)  # exclude df_nls
  }
  else {
    predictionModels <- rbind(df_gam, df_nls, df_lm, df_pwr, df_log, df_exp)
  }
  
  predictionModels <- predictionModels %>% 
    mutate(n = nrow(dataframe),
           PredictionType = case_when(knownX >= min(dataXcol) & knownX <= max(dataXcol) ~ "Interpolated",
                                    T ~ "Extrapolated"),
           ExtrapLength = case_when(PredictionType == "Extrapolated" ~ min(abs(min(dataXcol)-knownX), abs(max(dataXcol)-knownX)),
                                          T ~ 0)
    ) %>% 
    arrange(
      if(all(PredictionType == "Extrapolated") & !is.null(preferExtrapModel)){
        # if there is a preferred model when prediction is extrapolated, put that model first
        match(ModelName, c(preferExtrapModel, ModelName[ModelName != preferExtrapModel]))
      }else {
        # Otherwise, sort by SER
        # make sure SER is real number. When n=2, SER can be NA or Inf. If SER is NA or Inf, use RMSE instead
        if(all(is.finite(SER))) SER  else RMSE
      }
    )
  
  return(predictionModels)
}

# EXAMPLE CODE TO USE MODEL FOR REGRESSION
# predictionModel_final <- predictionModels %>% 
#   filter(SER == min(SER))
# 
# model_predictOrig <- tibble(fish_Stdz350=seq(0.05,max(prediction2000_2019$fish_Stdz350), by=0.005))
# model_predictOrig <- model_predictOrig %>% 
#   mutate(aq_Samp=predict(predictionModel_final$Model[[1]], model_predictOrig %>% select(fish_Stdz350)))  
